/*************************************************************************
 *
 *  C-to-C Translator
 *
 *  Rob Miller
 * 
 *************************************************************************/
/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <ctype.h>
#include "ast.h"
GLOBAL SUEtype *make_SUE(NodeType typ, const char *name, List *fields);

/* CilkPrefix: prefix for all preprocessor-generated symbols */
GLOBAL const char *CilkPrefix = "_cilk_";

PRIVATE char CilkNameBuf[100];
PRIVATE const char *ApplyPrefix(const char *root)
{
     sprintf(CilkNameBuf, "%s%s", CilkPrefix, root);
     return UniqueString(CilkNameBuf);
}

PRIVATE const char *MakeSyncName(int n)
{
     sprintf(CilkNameBuf, "%s%s%d", CilkPrefix, "sync", n);
     return (UniqueString(CilkNameBuf));
}
PRIVATE const char *MakeFrameTypeName(const char *proc)
{
     sprintf(CilkNameBuf, "%s%s_frame", CilkPrefix, proc);
     return (UniqueString(CilkNameBuf));
}
PRIVATE const char *MakeSlowProcName(const char *proc)
{
     sprintf(CilkNameBuf, "%s%s_slow", CilkPrefix, proc);
     return (UniqueString(CilkNameBuf));
}
PRIVATE const char *MakeScopeName(int n)
{
     sprintf(CilkNameBuf, "scope%d", n);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeLinkName(const char *proc)
{
     sprintf(CilkNameBuf, "%s%s_sig", CilkPrefix, proc);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeTempName(void)
{
     static int temp_count = 0;
     sprintf(CilkNameBuf, "%s%s%d", CilkPrefix, "temp", temp_count++);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeInletName(const char *proc, const char *inlet)
{
     sprintf(CilkNameBuf, "%s%s_%s_inlet", CilkPrefix, proc, inlet);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeImplicitInletName(const char *proc, int entry)
{
     sprintf(CilkNameBuf, "%s%s_%d_inlet", CilkPrefix, proc, entry);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeInletArgsTypeName(const char *inlet)
{
     sprintf(CilkNameBuf, "%s_args", inlet);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *FrameName, *HeaderName, *DummyReturnName;
PRIVATE const char *WorkerStateTypeName, *WorkerStateName;
PRIVATE const char *EntryName, *IndexName, *SizeName, *FunctionName;
PRIVATE const char *SynchronizeName, *SynchedName;
PRIVATE const char *GenericFrameTypeName;
PRIVATE const char *InitFrameName, *PushFrameName;
PRIVATE const char *XPopFrameResultName, *XPopFrameNoResultName;
PRIVATE const char *SetResultName, *SetNoResultName;
PRIVATE const char *InletArgsName, *InletResultName;
PRIVATE const char *AbortSlowName, *AbortStandaloneName;
PRIVATE const char *StartThreadFastName, *StartThreadSlowName,
     *BeforeSpawnFastName, *BeforeSpawnSlowName, *AfterSpawnFastName,
     *AfterSpawnSlowName, *AfterSyncSlowName, *BeforeSyncSlowName,
     *AtSyncFastName, *BeforeReturnFastName, *BeforeReturnSlowName,
     *BeforeReturnInletName, *AtThreadBoundarySlowName;
PRIVATE const char *OffsetofName;

PRIVATE void InitCilkTransform(void)
{
     /* frame names */
     FrameName = ApplyPrefix("frame");

     /* parts-of-frame names */
     HeaderName = UniqueString("header");
     DummyReturnName = UniqueString("dummy_return");

     /* parts of header */
     EntryName = UniqueString("entry");
     IndexName = UniqueString("index");
     SizeName = UniqueString("size");
     FunctionName = UniqueString("fn");

     /* parts of RTS */
     GenericFrameTypeName = UniqueString("CilkStackFrame");
     WorkerStateTypeName = UniqueString("CilkWorkerState");

     /* name of worker state */
     WorkerStateName = ApplyPrefix("ws");

     /* RTS macros */
     InitFrameName = UniqueString("CILK2C_INIT_FRAME");
     PushFrameName = UniqueString("CILK2C_PUSH_FRAME");
     XPopFrameResultName = UniqueString("CILK2C_XPOP_FRAME_RESULT");
     XPopFrameNoResultName = UniqueString("CILK2C_XPOP_FRAME_NORESULT");
     SetResultName = UniqueString("CILK2C_SET_RESULT");
     SetNoResultName = UniqueString("CILK2C_SET_NORESULT");
     SynchronizeName = UniqueString("CILK2C_SYNC");
     SynchedName = UniqueString("CILK2C_SYNCHED");
     AbortSlowName = UniqueString("CILK2C_ABORT_SLOW");
     AbortStandaloneName = UniqueString("CILK2C_ABORT_STANDALONE");
     StartThreadFastName = UniqueString("CILK2C_START_THREAD_FAST");
     StartThreadSlowName = UniqueString("CILK2C_START_THREAD_SLOW");
     BeforeSpawnFastName = UniqueString("CILK2C_BEFORE_SPAWN_FAST");
     BeforeSpawnSlowName = UniqueString("CILK2C_BEFORE_SPAWN_SLOW");
     AfterSpawnFastName = UniqueString("CILK2C_AFTER_SPAWN_FAST");
     AfterSpawnSlowName = UniqueString("CILK2C_AFTER_SPAWN_SLOW");
     AtSyncFastName = UniqueString("CILK2C_AT_SYNC_FAST");
     AtThreadBoundarySlowName = UniqueString("CILK2C_AT_THREAD_BOUNDARY_SLOW");
     BeforeSyncSlowName = UniqueString("CILK2C_BEFORE_SYNC_SLOW");
     AfterSyncSlowName = UniqueString("CILK2C_AFTER_SYNC_SLOW");
     BeforeReturnFastName = UniqueString("CILK2C_BEFORE_RETURN_FAST");
     BeforeReturnSlowName = UniqueString("CILK2C_BEFORE_RETURN_SLOW");
     BeforeReturnInletName = UniqueString("CILK2C_BEFORE_RETURN_INLET");
     OffsetofName = UniqueString("CILK_OFFSETOF");

     InletArgsName = ApplyPrefix("inletargs");
     InletResultName = ApplyPrefix("inletresult");
}

/* true iff transforming a cilk procedure */
PRIVATE int in_procedure = 0;
PRIVATE int in_inlet = 0;

PRIVATE Node *curr_proc;
PRIVATE const char *proc_name;
PRIVATE int sync_count;
typedef enum {
     Fast,			/* generating fast version of procedure */
     Slow,			/* generating slow version of procedure */
     Link,			/* generating linkage information       */
     InletProc			/* generating stand-alone inlet         */
} VersionType;
VersionType version;
List *link_initializers;
List *standalone_inlets;

PRIVATE Node *MakeFrameType(const char *procname)
{
     return MakeSdcl(EMPTY_TQ,
		     make_SUE(Sdcl, MakeFrameTypeName(procname), NULL));
}

PRIVATE Node *MakeGenericFrameType(void)
{
     return MakeTdef(EMPTY_TQ, GenericFrameTypeName);
}

PRIVATE Node *MakeWorkerStateType(void)
{
     return MakeTdef(EMPTY_TQ, WorkerStateTypeName);
}

PRIVATE Node *MakeInletArgsType(const char *inletname)
{
     return MakeSdcl(EMPTY_TQ,
		 make_SUE(Sdcl, MakeInletArgsTypeName(inletname), NULL));
}

PRIVATE Bool IsInletProc(Node *proc)
{
     Node *decl, *fdcl;
     assert(proc->typ == Proc);
     decl = proc->u.proc.decl;
     assert(decl->typ == Decl);
     fdcl = decl->u.decl.type;
     assert(fdcl->typ == Fdcl);
     return (TYPE_QUALS(fdcl->u.fdcl.tq) & T_INLET) ? 1 : 0;
}

/*
 * Figures out the type (Fdcl) of a spawned function.  Pass the
 * function expression (p->u.spawn.name) to this routine.
 * Also works for inlet decls.
 */
PRIVATE Node *GetSpawnFdcl(Node *spawned_fn)
{
     Node *type = NodeDataType(spawned_fn);

     if (IsPtrToFunction(type))
	  type = type->u.ptr.type;

     assert(type->typ == Fdcl);
     return type;
}

PRIVATE Node *TransformNode(Node *node);

PRIVATE List *TransformList(List *list)
{
     List *aptr;

     for (aptr = list; aptr; aptr = Rest(aptr)) {
	  Node *item = FirstItem(aptr);

	  SetItem(aptr, TransformNode(item));
     }

     return list;
}

PRIVATE inline Node *TransformConst(Node *node, ConstNode *u)
{
     return node;
}

PRIVATE inline Node *TransformId(Node *node, idNode *u)
{
     const char *scope;

     if (!u->decl)
	  return node;

     /* rename id to decl's name: implements static variable renaming */
     assert(u->decl->typ == Decl);
     u->text = u->decl->u.decl.name;

     /* get scope name of id's declaration */
     scope = u->decl->u.decl.scope;

     switch (version) {
	 case Fast:
	      if (u->decl->u.decl.kind == 2) {
		   Node *result = MakeBinop('.',
					    MakeBinop(ARROW,
						      MakeId(FrameName),
						      MakeId(scope)),
					    MakeId(u->decl->u.decl.name));
		   SetCoords(result, node->coord, Subtree);
		   return (result);
	      }
	      break;
	 case Slow:
	      if (u->decl->u.decl.kind == 2 || u->decl->u.decl.kind == 3) {
		   Node *result = MakeBinop('.',
					    MakeBinop(ARROW,
						      MakeId(FrameName),
						      MakeId(scope)),
					    MakeId(u->decl->u.decl.name));
		   SetCoords(result, node->coord, Subtree);
		   return (result);
	      }
	      break;
	 case Link:
	      return MakeBinop('.',
			       MakeId(scope),
			       MakeId(u->decl->u.decl.name));
	 case InletProc:
	      if (u->decl->u.decl.kind == 4) {
		   Node *result = MakeBinop(ARROW,
					    MakeId(InletArgsName),
					    MakeId(u->decl->u.decl.name));
		   SetCoords(result, node->coord, Subtree);
		   return result;
	      } else if (u->decl->u.decl.kind != 0) {
		   Node *result = MakeBinop('.',
					    MakeBinop(ARROW,
						      MakeId(FrameName),
						      MakeId(scope)),
					    MakeId(u->decl->u.decl.name));
		   SetCoords(result, node->coord, Subtree);
		   return result;
	      }
	      break;
     }
     return node;
}

PRIVATE inline Node *TransformBinop(Node *node, binopNode *u)
{
     u->left = TransformNode(u->left);
     u->right = TransformNode(u->right);

     return node;
}

PRIVATE inline Node *TransformUnary(Node *node, unaryNode *u)
{
     if (version == Link && u->op == INDIR) {
	  /* 
	   * transform *a to a[0], *(a+n) to a[n], *(n+a) to a[n], 
	   * and *&x to x 
	   */
	  switch (u->expr->typ) {
	      case Id:
		   return (MakeArray(TransformNode(u->expr), List1(MakeConstSint(0))));
	      case Binop:
		   if (u->expr->u.binop.op == '+' ||
		       u->expr->u.binop.op == '-') {
			if (NodeIsConstant(u->expr->u.binop.left))
			     return (MakeArray(TransformNode(u->expr->u.binop.right),
					       List1(TransformNode(u->expr->u.binop.left))));
			if (NodeIsConstant(u->expr->u.binop.right))
			     return (MakeArray(TransformNode(u->expr->u.binop.left),
					       List1(TransformNode(u->expr->u.binop.right))));
		   }
		   break;
	      case Unary:
		   if (u->expr->u.unary.op == ADDRESS)
			return (TransformNode(u->expr->u.unary.expr));
		   break;
	      default:
		   break;
	  }
     }
     u->expr = TransformNode(u->expr);

     return node;
}

PRIVATE inline Node *TransformCast(Node *node, castNode *u)
{
     u->type = TransformNode(u->type);
     u->expr = TransformNode(u->expr);

     return node;
}

PRIVATE inline Node *TransformComma(Node *node, commaNode *u)
{
     u->exprs = TransformList(u->exprs);
     return node;
}

PRIVATE inline Node *TransformTernary(Node *node, ternaryNode *u)
{
     u->cond = TransformNode(u->cond);
     u->true = TransformNode(u->true);
     u->false = TransformNode(u->false);

     return node;
}

PRIVATE inline Node *TransformArray(Node *node, arrayNode *u)
{
     u->name = TransformNode(u->name);
     u->dims = TransformList(u->dims);

     return node;
}

PRIVATE inline Node *TransformCall(Node *node, callNode *u)
{
     u->name = TransformNode(u->name);
     u->args = TransformList(u->args);

     return node;
}

PRIVATE inline Node *TransformInitializer(Node *node, initializerNode *u)
{
     u->exprs = TransformList(u->exprs);
     return node;
}

PRIVATE inline Node *TransformImplicitCast(Node *node, implicitcastNode *u)
{
     u->expr = TransformNode(u->expr);
     return node;
}

/*************************************************************************/
/*                                                                       */
/*                          Statement nodes                              */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Node *TransformLabel(Node *node, labelNode *u)
{
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformSwitch(Node *node, SwitchNode *u)
{
     u->expr = TransformNode(u->expr);
     u->stmt = TransformNode(u->stmt);

     return node;
}

PRIVATE inline Node *TransformCase(Node *node, CaseNode *u)
{
     u->stmt = TransformNode(u->stmt);

     return node;
}

PRIVATE inline Node *TransformDefault(Node *node, DefaultNode *u)
{
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformIf(Node *node, IfNode *u)
{
     u->expr = TransformNode(u->expr);
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformIfElse(Node *node, IfElseNode *u)
{
     u->expr = TransformNode(u->expr);
     u->true = TransformNode(u->true);
     u->false = TransformNode(u->false);
     return node;
}

PRIVATE inline Node *TransformWhile(Node *node, WhileNode *u)
{
     u->expr = TransformNode(u->expr);
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformDo(Node *node, DoNode *u)
{
     u->expr = TransformNode(u->expr);
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformFor(Node *node, ForNode *u)
{
     u->cond = TransformNode(u->cond);
     u->init = TransformNode(u->init);
     u->next = TransformNode(u->next);
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformGoto(Node *node, GotoNode *u)
{
     return node;
}

PRIVATE inline Node *TransformContinue(Node *node, ContinueNode *u)
{
     return node;
}

PRIVATE inline Node *TransformBreak(Node *node, BreakNode *u)
{
     return node;
}

PRIVATE inline Node *TransformReturn(Node *node, ReturnNode *u)
{
     Node *result;
     List *statements = NULL;
     List *decls = NULL;

     if (!in_procedure)
	  return node;

     if (u->needs_sync) {
	  result = MakeBlockCoord(NULL,
				  NULL,
				  List2(MakeSyncCoord(node->coord),
				  MakeReturnCoord(u->expr, node->coord)),
				  node->coord,
				  UnknownCoord);
	  /* copy over analysis results from return to sync node */
	  ((Node *) FirstItem(result->u.Block.stmts))->analysis.livevars = u->livevars;
	  ((Node *) FirstItem(result->u.Block.stmts))->analysis.dirtyvars = u->dirtyvars;
	  return TransformNode(result);
     }
     u->expr = TransformNode(u->expr);

     /* I'm not sure if this condition is right... -KHR */
     if (in_inlet && (version == Fast || version == Slow))
	  return node;
     else if (version == Fast) {
	  /* return x ==> { T _tmp = x; _BEFORE_RETURN_FAST(); return _tmp; } */
	  /* return ==> { _BEFORE_RETURN_FAST(); return; } */
	  /* if u->needs_sync == 1 then no need to create temparary value 
	   * as it is already done in AnalyzeReturnExpressions. -fengmd */
	  if (u->expr && (!(u->needs_sync))) {
	       const char *tmp = MakeTempName();
	       Node *returntype = curr_proc->u.proc.decl->u.decl.type->u.fdcl.returns;
	       Node *decl = MakeDecl(tmp,
				     EMPTY_TQ,
				     NodeCopy(returntype, Subtree),
				     u->expr,
				     NULL);
	       decls = List1(decl);

	       statements = List2(MakeCall(MakeId(BeforeReturnFastName), NULL),
				  MakeReturn(MakeId(tmp)));
	  } else
	       statements = List2(MakeCall(MakeId(BeforeReturnFastName), NULL),
				  node);
     } else if (version == Slow) {
	  /* return x  ==>  {_SET_RESULT(x); _SET_VALID; return;} */
	  /* return  ==>  {_SET_VALID; return;} */
	  if (u->expr)
	       statements = List1(MakeCall(MakeId(SetResultName),
					   List1(u->expr)));
	  else
	       statements = List1(MakeCall(MakeId(SetNoResultName), NULL));

	  statements = AppendItem(statements,
			   MakeCall(MakeId(BeforeReturnSlowName), NULL));
	  statements = AppendItem(statements, MakeReturn(NULL));
     } else if (version == InletProc) {
	  /* return x  ==>  {*_inletresult = x; return;} */
	  /* return  ==>  return */
	  if (!u->expr)
	       return MakeBlock(NULL, NULL,
			    List2(MakeCall(MakeId(BeforeReturnInletName),
					   NULL),
				  node));
	  statements = List3(MakeBinop('=',
				       MakeUnary('*',
						 MakeId(InletResultName)),
				       u->expr),
			   MakeCall(MakeId(BeforeReturnInletName), NULL),
			     MakeReturn(NULL));
     } else
	  FAIL("bad version in TransformReturn");

     result = MakeBlock(NULL, decls, statements);
     SetCoords(result, node->coord, Subtree);
     return result;
}

PRIVATE inline Node *TransformBlock(Node *node, BlockNode *u)
{
     ListMarker decl_marker;
     Node *decl;

     IterateList(&decl_marker, u->decl);
     while (NextOnList(&decl_marker, (GenericREF) & decl)) {
	  decl = TransformNode(decl);
	  SetCurrentOnList(&decl_marker, decl);
     }

     u->stmts = TransformList(u->stmts);

     return node;
}

/*************************************************************************/
/* Determine whether a declaration has the CONST type qualifier          */
/*************************************************************************/
PRIVATE Bool HasConstQual(Node *node)
{
     return (NodeIsConstQual(NodeDataType(node)) ||
	     NodeIsConstQual(NodeDataTypeSuperior(node)));
}

/*************************************************************************/
/*                                                                       */
/*                             Type nodes                                */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Node *TransformPrim(Node *node, primNode *u)
{
     return node;
}

PRIVATE inline Node *TransformTdef(Node *node, tdefNode *u)
{
     return node;
}

PRIVATE inline Node *TransformPtr(Node *node, ptrNode *u)
{
     u->type = TransformNode(u->type);
     return node;
}

PRIVATE inline Node *TransformAdcl(Node *node, adclNode *u)
{
     u->type = TransformNode(u->type);
     u->tq = NodeTq(u->type);
     u->dimp->dim = TransformNode(u->dimp->dim);
     return node;
}

PRIVATE inline Node *TransformFdcl(Node *node, fdclNode *u)
{
     u->args = TransformList(u->args);
     u->returns = TransformNode(u->returns);

     if (u->tq & T_PROCEDURE) {
	  List *args;
	  Node *n;
	  args = u->args;
	  n = MakeDecl(WorkerStateName,
		       T_FORMAL_DECL,
		       MakePtr(T_CONST, MakeWorkerStateType()),
		       NULL,
		       NULL);

	  if (args) {
	       if (IsVoidType(FirstItem(args))) {
		    u->args = MakeNewList(n);
	       } else {
		    u->args = ConsItem(n, args);
	       }
	  } else {
	       /* old-style fdecl, don't do anything and hope for the best */
	  }
     }

     return node;
}


#define UnionCheckFields(x) x=x
#define AssignEnumValues(x) x=x

PRIVATE void TransformSUE(SUEtype *sue)
{
     if (sue->transformed == FALSE) {
	  ListMarker marker;
	  Node *field;

	  /* To stop infinite recursion */
	  sue->transformed = TRUE;

	  /* Loop over the fields of the SDCL */
	  IterateList(&marker, sue->fields);
	  while (NextOnList(&marker, (GenericREF) & field)) {
	       assert(field->typ == Decl);
	       field->u.decl.type = TransformNode(field->u.decl.type);
	  }
     }
}


PRIVATE inline Node *TransformSdcl(Node *node, sdclNode *u)
{
     TransformSUE(u->type);
     return node;
}

PRIVATE inline Node *TransformUdcl(Node *node, udclNode *u)
{
     TransformSUE(u->type);
     return node;
}

PRIVATE inline Node *TransformEdcl(Node *node, edclNode *u)
{
     TransformSUE(u->type);
     return node;
}

/*************************************************************************/
/*                                                                       */
/*                      Other nodes (decls et al.)                       */
/*                                                                       */
/*************************************************************************/
PRIVATE inline Node *TransformDecl(Node *node, declNode *u)
{
     u->type = TransformNode(u->type);
     u->init = TransformNode(u->init);
     u->bitsize = TransformNode(u->bitsize);

     if (in_procedure && (u->kind == 2 ||
			  (version == Slow && u->kind == 3))) {
	  if (!u->init)
	       return NULL;
	  {
	       /* We do an incredible hack to get a frame-only variable
	        * with initialization to compile correctly.
	        *     t x = i;
	        * compiles to:
	        *     int tmp1 = ({t tmp2 = i; f->x = tmp2; tmp1;});
	        */
	       const char *tmp1 = MakeTempName();
	       const char *tmp2 = MakeTempName();
	       Node *newinit, *newdecl, *assign;
	       Node *id = MakeId(u->name);
	       id->u.id.decl = node;

	       if (NodeDataType(node)->typ == Adcl)
		    assign = MakeCall(MakeId(UniqueString("memcpy")),
				      List3(TransformNode(id),
					    MakeId(tmp2),
					    MakeUnary(SIZEOF,
						      MakeId(tmp2))));
	       else
		    assign = MakeBinop('=', TransformNode(id), MakeId(tmp2));

	       newinit = MakeBlock(NULL,
				   List1(MakeDecl(tmp2,
						  u->tq,
						  u->type,
						  u->init,
						  u->bitsize)),
				   List2(assign,
					 MakeId(tmp1)));
	       newdecl = MakeDecl(tmp1,
				  EMPTY_TQ,
				  MakePrim(EMPTY_TQ, Sint),
				  newinit,
				  NULL);
	       SetCoords(newdecl, node->coord, Subtree);
	       return newdecl;
	  }
     }
     /* delete static declarations (they have been globalized) -KHR */
     if (in_procedure && NodeStorageClass(node) == T_STATIC)
	  return NULL;

     /*
      * remove const in slow declaration (to allow it to be loaded from 
      * frame) 
      */
     if (in_procedure && version == Slow)
	  NodeRemoveTq(node->u.decl.type, T_CONST);

     /*
      * Generate attribute for T_SHARED and T_PRIVATE variables
      */
     if (NodeDeclQuals(node) & T_PRIVATE) {
	  if (GenerateSegments) {
	       if (!HasConstQual(node))
		    u->attribs =
			 ConsItem(
			      MakeAttrib(
				   UniqueString("CILK_ATTRIBUTE_PRIVATE_VAR"),
				   NULL),
			      u->attribs);
	  } else
	       WarningCoord(2, node->coord,
		   "`private' qualifier not supported on this target\n");
     } else if (NodeDeclQuals(node) & T_SHARED ||
		(GenerateSegments &&
		 NodeDeclLocation(node) == T_TOP_DECL &&
		 NodeStorageClass(node) != T_EXTERN &&
		 NodeStorageClass(node) != T_REGISTER &&
		 NodeStorageClass(node) != T_TYPEDEF &&
		 NodeDataType(node)->typ != Fdcl)) {
	  if ((NodeDeclQuals(node) & T_SHARED) && !GenerateSegments) {
#if 0	       
	       WarningCoord(2, node->coord,
		    "`shared' qualifier not supported on this target\n");
#endif
	  } else if (!HasConstQual(node))
	       u->attribs = 
		    ConsItem(
			 MakeAttrib(UniqueString("CILK_ATTRIBUTE_SHARED_VAR"),
				    NULL),
			 u->attribs);
     }

     return node;
}

PRIVATE inline Node *TransformAttrib(Node *node, attribNode *u)
{
     return node;
}

PRIVATE void MakeInletProc(Node *node)
{
     Node *decl = node->u.proc.decl;
     const char *name = MakeInletName(proc_name, decl->u.decl.name);
     Node *args_decl, *args_type;
     Node *inlet_type;
     Node *inlet_decl;
     Node *inlet;

     if (!IsVoidType(FirstItem(decl->u.decl.type->u.fdcl.args))) {
	  args_decl = MakeSdcl(EMPTY_TQ | T_SUE_ELABORATED,
			       make_SUE(Sdcl,
					MakeInletArgsTypeName(name),
					decl->u.decl.type->u.fdcl.args));
	  args_type = MakeInletArgsType(name);
     } else {
	  args_decl = NULL;
	  args_type = MakePrim(EMPTY_TQ, Void);
     }

     inlet_type = MakeFdcl(EMPTY_TQ,
			   List4(MakeDecl(WorkerStateName,
					  T_FORMAL_DECL,
					  MakePtr(T_CONST,
						  MakeWorkerStateType()),
					  NULL,
					  NULL),
				 MakeDecl(FrameName,
					  EMPTY_TQ,
					  MakePtr(EMPTY_TQ,
					       MakeFrameType(proc_name)),
					  NULL,
					  NULL),
				 MakeDecl(InletArgsName,
					  EMPTY_TQ,
					  MakePtr(EMPTY_TQ,
						  args_type),
					  NULL,
					  NULL),
				 MakeDecl(InletResultName,
					  EMPTY_TQ,
					  MakePtr(EMPTY_TQ,
			      NodeCopy(decl->u.decl.type->u.fdcl.returns,
				       Subtree)),
					  NULL,
					  NULL)),
			   MakePrim(EMPTY_TQ, Void));

     inlet_decl = MakeDecl(name,
			   T_STATIC,
			   inlet_type,
			   NULL,
			   NULL);
     if (args_decl)
	  SetCoords(args_decl, node->coord, Subtree);
     SetCoords(inlet_decl, node->coord, Subtree);
     in_inlet = 1;
     version = InletProc;
     inlet = MakeProc(inlet_decl,
		      TransformNode(node->u.proc.body));
     version = Slow;
     in_inlet = 0;
     standalone_inlets = AppendItem(standalone_inlets, args_decl);
     standalone_inlets = AppendItem(standalone_inlets, inlet);
}

PRIVATE void MakeImplicitInletProc(Node *node)
{
     const char *name = MakeImplicitInletName(proc_name, sync_count + 1);	/* +1 because inlet is made before entry number is allocated */
     Node *inlet_type, *inlet_decl, *stmt, *inlet;
     Node *rhs_type, *lhs_type;

     assert(node->typ == Spawn);
     assert(node->u.spawn.assign_op != '=');
     /*
      * Generate inlet code for implicit inlet.  It should look
      * something like this:
      * void _foo_inlet(_foo_frame *_frame, int *arg, int *result)
      * { *result += *arg; }
      * Note: frame pointer is not used.
      */
     rhs_type = NodeCopy(GetSpawnFdcl(node->u.spawn.name)->u.fdcl.returns, Subtree);
     lhs_type = NodeCopy(NodeDataType(node->u.spawn.receiver), Subtree);

     inlet_type = MakeFdcl(EMPTY_TQ,
			   List4(MakeDecl(WorkerStateName,
					  T_FORMAL_DECL,
					  MakePtr(T_CONST,
						  MakeWorkerStateType()),
					  NULL,
					  NULL),
				 MakeDecl(FrameName,
					  EMPTY_TQ,
					  MakePtr(EMPTY_TQ,
					       MakeFrameType(proc_name)),
					  NULL,
					  NULL),
				 MakeDecl(InletArgsName,
					  EMPTY_TQ,
					  MakePtr(EMPTY_TQ,
						  rhs_type),
					  NULL,
					  NULL),
				 MakeDecl(InletResultName,
					  EMPTY_TQ,
					  MakePtr(EMPTY_TQ,
						  lhs_type),
					  NULL,
					  NULL)),
			   MakePrim(EMPTY_TQ, Void));

     inlet_decl = MakeDecl(name,
			   T_STATIC,
			   inlet_type,
			   NULL,
			   NULL);
     stmt = MakeBinop(node->u.spawn.assign_op,
		      MakeUnary('*', MakeId(InletResultName)),
		      MakeUnary('*', MakeId(InletArgsName)));
     inlet = MakeProc(inlet_decl,
		      MakeBlock(NULL,
				NULL,
				List1(stmt)));

     SetCoords(inlet, UnknownCoord, Subtree);
     standalone_inlets = AppendItem(standalone_inlets, inlet);
}

PRIVATE inline Node *TransformProc(Node *node, procNode *u)
{
     Node *standalone_copy = NULL;

     if (IsInletProc(node)) {
	  assert(in_procedure && !in_inlet);
	  if (version == Slow)
	       standalone_copy = NodeCopy(node, Subtree);
	  in_inlet = 1;
     }

     u->decl = TransformNode(u->decl);
     u->body = TransformNode(u->body);

     if (IsInletProc(node)) {
	  in_inlet = 0;

	  /* make standalone inlet */
	  if (version == Slow)
	       MakeInletProc(standalone_copy);
     }
     return node;
}

PRIVATE inline Node *TransformText(Node *node, textNode *u)
{
     return node;
}

PRIVATE List *SaveVariables(List *vars, Node *except)
{
     ListMarker m;
     Node *var;
     List *statements = NULL;

     IterateList(&m, vars);
     while (NextOnList(&m, (GenericREF) & var)) {
	  if (var != except &&
	      (var->u.decl.kind == 1 ||
	       (version == Fast && var->u.decl.kind == 3))) {
	       statements = AppendItem(statements,
				       MakeBinop('=',
						 MakeBinop('.',
							 MakeBinop(ARROW,
						       MakeId(FrameName),
					      MakeId(var->u.decl.scope)),
					       MakeId(var->u.decl.name)),
					      MakeId(var->u.decl.name)));
	  }
     }
     return (statements);
}

PRIVATE List *RestoreVariables(List *vars)
{
     ListMarker m;
     Node *var;
     List *statements = NULL;

     IterateList(&m, vars);
     while (NextOnList(&m, (GenericREF) & var)) {
	  if ((var->u.decl.kind == 1 ||
	       (version == Fast && var->u.decl.kind == 3))) {
	       statements = AppendItem(statements,
				       MakeBinop('=',
						 MakeId(var->u.decl.name),
						 MakeBinop('.',
							 MakeBinop(ARROW,
						       MakeId(FrameName),
					      MakeId(var->u.decl.scope)),
					     MakeId(var->u.decl.name))));
	  }
     }
     return (statements);
}

/*
 * This function generates one of two macros
 *
 * _XPOP_FRAME_RESULT(frame, return_value, stuff for slow, result)
 * _XPOP_FRAME_NORESULT(frame, return_value, stuff for slow)
 *
 * depending on the presence of a receiver for the result.
 */

PRIVATE inline Node *MakeXPopFrame(Node *node, spawnNode *u)
{
     Node *frame;
     Node *return_value;

     /* return type of the procedure being compiled, not the procedure *
      * being spawned! */
     Node *returntype = curr_proc->u.proc.decl->u.decl.type->u.fdcl.returns;

     frame = MakeId(FrameName);

     if (version == Slow)
	  return_value = MakeText(UniqueString("/* return nothing */"),
				  FALSE);
     else if (IsScalarType(returntype))
	  return_value = MakeConstSint(0);
     else if (IsVoidType(returntype))
	  return_value = MakeText(UniqueString("/* return nothing */"),
				  FALSE);
     else
	  return_value = MakeBinop(ARROW,
				   MakeId(FrameName),
				   MakeId(DummyReturnName));

     if (u->receiver)
	  return MakeCall(MakeId(XPopFrameResultName),
			  List3(frame,
				return_value,
				u->receiver));
     else
	  return MakeCall(MakeId(XPopFrameNoResultName),
			  List2(frame,
				return_value));
}

PRIVATE Node *MakeSetIndex(Node *receiver_addr)
{
     /* sets index of current frame to be the index of receiver in the *
      * frame. */
     /* _frame->_header.index = (char*) receiver_addr - (char*)_frame */

     return MakeBinop('=',
		      MakeBinop('.',
				MakeBinop(ARROW,
					  MakeId(FrameName),
					  MakeId(HeaderName)),
				MakeId(IndexName)),
		      MakeBinop('-',
				MakeCast(MakePtr(EMPTY_TQ,
					       MakePrim(EMPTY_TQ, Char)),
					 receiver_addr),
				MakeCast(MakePtr(EMPTY_TQ,
					       MakePrim(EMPTY_TQ, Char)),
					 MakeId(FrameName))));
}

PRIVATE Bool SimplePointer(Node *receiver);

/* returns TRUE if receiver's address in the frame can be calculated
 * statically, and if receiver is side-effect-free.  Returns FALSE otherwise.
 */
PRIVATE Bool SimpleReceiver(Node *receiver)
{
     List *dims;

     switch (receiver->typ) {
	 case Id:
	      if (receiver->u.id.decl->u.decl.kind != 0)
		   return TRUE;
	      break;
	 case Binop:
	      if (receiver->u.binop.op == '.')
		   return SimpleReceiver(receiver->u.binop.left);
	      break;
	 case Unary:
	      if (receiver->u.unary.op == INDIR)
		   return SimplePointer(receiver->u.unary.expr);
	      break;
	 case Array:
	      /* simple if all array dimensions are constant */
	      for (dims = receiver->u.array.dims; dims; dims = Rest(dims))
		   if (!NodeIsConstant(FirstItem(dims)))
			return FALSE;
	      return SimplePointer(receiver->u.array.name);
	 default:
	      /* there are probably other cases, but I don't feel like *
	       * dealing * with them now. KHR */
	      break;
     }
     return FALSE;
}
/* returns TRUE if pointer's offset in frame can be calculated statically,
 * and if receiver is side-effect-free.  Returns FALSE otherwise.
 */
PRIVATE Bool SimplePointer(Node *pointer)
{
     switch (pointer->typ) {
	 case Id:
	      if (pointer->u.id.decl->u.decl.type->typ == Adcl)
		   return TRUE;
	      break;
	 case Binop:
	      if (pointer->u.binop.op == '+' || pointer->u.binop.op == '-') {
		   if (NodeIsConstant(pointer->u.binop.right) &&
		       SimplePointer(pointer->u.binop.left))
			return TRUE;
		   if (NodeIsConstant(pointer->u.binop.left) &&
		       SimplePointer(pointer->u.binop.right))
			return TRUE;
	      }
	      break;
	 case Unary:
	      if (pointer->u.unary.op == ADDRESS)
		   return SimpleReceiver(pointer->u.unary.expr);
	      break;
	 case Cast:
	      return SimplePointer(pointer->u.cast.expr);
	 case ImplicitCast:
	      return SimplePointer(pointer->u.implicitcast.expr);
	 default:
	      break;
     }
     return FALSE;
}

/* gcc asm extensions */

PRIVATE inline Node *TransformAsm(Node *node, asmNode *u)
{
     u->template = TransformNode(u->template);
     u->output = TransformList(u->output);
     u->input = TransformList(u->input);
     u->clobbered = TransformList(u->clobbered);
     return node;
}
PRIVATE inline Node *TransformAsmArg(Node *node, asmargNode *u)
{
     u->constraint = TransformNode(u->constraint);
     u->expr = TransformNode(u->expr);
     return node;
}

PRIVATE Bool side_effect_free;

PRIVATE void CheckSideEffectFree(Node *node)
{
     switch (node->typ) {
	 case Binop:
	      if (IsAssignmentOp(node->u.binop.op)) {
		   side_effect_free = FALSE;
		   return;
	      }
	      break;
	 case Unary:
	      if (IsIncDecOp(node->u.unary.op)) {
		   side_effect_free = FALSE;
		   return;
	      }
	      break;
	 case Block:
	 case Call:
	 case Spawn:
	      side_effect_free = FALSE;
	      return;
	 case Label:
	 case Switch:
	 case Case:
	 case Default:
	 case If:
	 case IfElse:
	 case While:
	 case Do:
	 case For:
	 case Goto:
	 case Continue:
	 case Break:
	 case Return:
	 case Sync:
	      FAIL("Statement in spawn arg");
	      break;
	 case Prim:
	 case Tdef:
	 case Adcl:
	 case Fdcl:
	 case Sdcl:
	 case Udcl:
	 case Edcl:
	 case Decl:
	 case Attrib:
	 case Proc:
	      /* don't need to touch types */
	      return;
	 default:
	      break;
     }

#define CHILD(n)   CheckSideEffectFree(n)
     ASTWALK(node, CHILD);
#undef CHILD
}

PRIVATE Bool SideEffectFree(Node *node)
{
     side_effect_free = TRUE;
     CheckSideEffectFree(node);
     return side_effect_free;
}

/* returns a decl of a temporary variable with type <type> */
PRIVATE Node *MakeTempDecl(Node *type)
{
     const char *temp_arg = MakeTempName();
     return MakeDecl(temp_arg,
		     EMPTY_TQ,
		     NodeCopy(type, Subtree),
		     NULL,
		     NULL);
}

PRIVATE inline List *TransformSpawnArgs(List *args,
					List **declarations,
					List **statements)
{
     List *side_effect_free_args;
     Node *arg, *temp_arg, *temp_decl;
     ListMarker marker;

     side_effect_free_args = NULL;
     IterateList(&marker, args);
     while (NextOnList(&marker, (GenericREF) & arg)) {
	  if (!SideEffectFree(arg)) {
	       temp_decl = MakeTempDecl(NodeDataType(arg));
	       temp_arg = MakeId(temp_decl->u.decl.name);

	       *declarations = AppendItem(*declarations,
					  temp_decl);
	       *statements = AppendItem(*statements,
					MakeBinop('=',
					  MakeId(temp_decl->u.decl.name),
						  TransformNode(arg)));
	  } else
	       temp_arg = TransformNode(arg);
	  side_effect_free_args = AppendItem(side_effect_free_args, temp_arg);
     }
     
     return side_effect_free_args;
}

PRIVATE Node *TransformReceiver(Node *receiver,
				OpType assign_op,
				Node *return_type,
				Node **index,	/* linkage info         */
				Node **assignment,	/* assignment statement */
				List **declarations,
				List **statements)
{
     Node *rec, *receiver_tmp, *receiver_addr;
     Node *spawn_result = NULL;
     int receiver_type;		/* 
				 * 0 - no receiver
				 * 1 - register variable
				 * 2 - fixed location in frame
				 *       (SimpleReceiver) 
				 * 3 - unknown location
				 *  or side-effecting receiver */
     /* figure out receiver type */
     if (!receiver)
	  receiver_type = 0;
     else if (receiver->typ == Id &&
	      (receiver->u.id.decl->u.decl.kind == 1 ||
	     (version == Fast && receiver->u.id.decl->u.decl.kind == 3)))
	  receiver_type = 1;
     else if (SimpleReceiver(receiver))
	  receiver_type = 2;
     else
	  receiver_type = 3;

     /* generate index part of link information - slow version only. */
     if (version == Slow)
	  switch (receiver_type) {
	      case 0:		/* no receiver */
		   *index = MakeConstSint(0);
		   break;
	      case 1:		/* register variable */
		   *index = MakeCall(MakeId(OffsetofName),
				     List2(MakeFrameType(proc_name),
					   MakeBinop('.',
			       MakeId(receiver->u.id.decl->u.decl.scope),
			     MakeId(receiver->u.id.decl->u.decl.name))));
		   break;
	      case 2:		/* fixed loc in frame */
		   rec = NodeCopy(receiver, Subtree);
		   version = Link;
		   rec = TransformNode(rec);
		   version = Slow;
		   *index = MakeCall(MakeId(OffsetofName),
				     List2(MakeFrameType(proc_name),
					   rec));
		   break;
	      case 3:		/* unknown location */
		   *index = MakeConstSint(-1);
		   break;
	      default:
		   FAIL("bad receiver type");
		   break;
	  }
     /* make <spawn_result>, <assignment> */
     switch (receiver_type) {
	 case 0:		/* no receiver */
	      spawn_result = NULL;
	      *assignment = NULL;
	      break;
	 case 1:		/* register variable */
	      if (assign_op == '=') {
		   spawn_result = receiver;
		   *assignment = NULL;
		   break;
	      }
	      /* otherwise, fall through */
	 case 2:		/* fixed loc in frame */
	      receiver_tmp = MakeTempDecl(NodeCopy(return_type, Subtree));
	      *declarations = AppendItem(*declarations, receiver_tmp);
	      spawn_result = MakeId(receiver_tmp->u.decl.name);
	      *assignment = MakeBinop(assign_op,
				      TransformNode(receiver),
				      MakeId(receiver_tmp->u.decl.name));
	      break;
	 case 3:		/* unknown location */
	      receiver_tmp = MakeTempDecl(NodeCopy(return_type, Subtree));
	      *declarations = AppendItem(*declarations, receiver_tmp);
	      receiver_addr = MakeTempDecl(MakePtr(EMPTY_TQ,
						NodeDataType(receiver)));
	      *declarations = AppendItem(*declarations, receiver_addr);
	      *statements = AppendItem(*statements,
				       MakeBinop('=',
				      MakeId(receiver_addr->u.decl.name),
					 TransformNode(MakeUnary(ADDRESS,
							    receiver))));
	      spawn_result = MakeId(receiver_tmp->u.decl.name);
	      *assignment = MakeBinop(assign_op,
				      MakeUnary('*',
				     MakeId(receiver_addr->u.decl.name)),
				      MakeId(receiver_tmp->u.decl.name));
	      *statements = AppendItem(*statements,
		       MakeSetIndex(MakeId(receiver_addr->u.decl.name)));
	      break;
     }
     return spawn_result;
}

PRIVATE void TransformSpawn2(Node *node, spawnNode *u,
			     List **declarations, List **statements)
{
     Node *index, *assignment;
     Node *worker_state;
     List *nargs;
     Node *return_type = GetSpawnFdcl(u->name)->u.fdcl.returns;

     assert(in_procedure && !in_inlet);

     /* reserve a sync number */
     sync_count++;

     u->receiver = TransformReceiver(u->receiver, u->assign_op, return_type, &index,
				  &assignment, declarations, statements);
     assert(SideEffectFree(u->name));	/* need this for now -KHR */
     u->name = TransformNode(u->name);
     u->args = TransformSpawnArgs(u->args, declarations, statements);

     /* add extra argument for worker state */
     worker_state = MakeId(WorkerStateName);
     nargs = ConsItem(worker_state, u->args);

     /* set up link information */
     if (version == Slow && !u->in_inlet) {
	  Node *size, *inlet, *argsize, *argindex;
	  if (u->receiver)
	       size = MakeUnary(SIZEOF, NodeCopy(return_type, Subtree));
	  else
	       size = MakeConstSint(0);

	  if (u->assign_op == '=') {
	       inlet = MakeConstSint(0);
	       argsize = MakeConstSint(0);
	       argindex = MakeConstSint(0);
	  } else {
	       inlet = MakeId(MakeImplicitInletName(proc_name, sync_count));
	       argsize = MakeUnary(SIZEOF, NodeCopy(return_type, Subtree));
	       argindex = MakeConstSint(0);
	  }

	  link_initializers =
	      AppendItem(link_initializers,
			 MakeInitializer(List5(size,
					       index,
					       inlet,
					       argsize,
					       argindex)));
     }
     /* make entry setting statement */
     *statements = AppendItem(*statements,
			      MakeBinop('=',
					MakeBinop('.',
						  MakeBinop(ARROW,
						       MakeId(FrameName),
						     MakeId(HeaderName)),
						  MakeId(EntryName)),
					MakeConstSint(sync_count)));

     /* save live, dirty variables */
     *statements = JoinLists(*statements,
			  SaveVariables(node->analysis.dirtyvars, NULL));

     /* Insert _BEFORE_SPAWN_[FAST|SLOW] hook */
     *statements =
	 AppendItem(*statements,
		  MakeCall(MakeId(version == Slow ? BeforeSpawnSlowName :
				  BeforeSpawnFastName),
			   NULL)
	 );

     /* push frame */
     *statements = AppendItem(*statements,
			      MakeCall(MakeId(PushFrameName),
				       List1(MakeId(FrameName))));

     /* do call */
     if (u->receiver)
	  *statements = AppendItem(*statements,
				   MakeBinop('=',
					     u->receiver,
					     MakeCall(u->name,
						      nargs)));
     else
	  *statements = AppendItem(*statements,
				   MakeCall(u->name,
					    nargs));

     /* Extended pop: pop, check for steal, locking (in some unspecified order) */
     *statements = AppendItem(*statements,
			      MakeXPopFrame(node, u));

     /* Insert _AFTER_SPAWN_[FAST|SLOW] hook */
     *statements =
	 AppendItem(*statements,
		    MakeCall(MakeId(version == Slow ? AfterSpawnSlowName :
				    AfterSpawnFastName),
			     NULL)
	 );

     /* save spawn result somewhere, if necessary */
     if (assignment != NULL)
	  *statements = AppendItem(*statements, assignment);
}

PRIVATE inline Node *TransformSpawn(Node *node, spawnNode *u)
{
     List *declarations = NULL;
     List *statements = NULL;
     Node *result;
     Coord coord = u->receiver ? u->receiver->coord : node->coord;

     /* make standalone implicit inlet */
     if (version == Slow && u->assign_op != '=')
	  MakeImplicitInletProc(node);

     /* do guts of transformation */
     TransformSpawn2(node, u, &declarations, &statements);

     /* Add entry recovery statements to slow version */
     if (version == Slow) {
	  statements = AppendItem(statements,
				  MakeIf(MakeConstSint(0),
					 MakeBlock(NULL, NULL,
			     ConsItem(MakeLabel(MakeSyncName(sync_count),
						NULL),
			   RestoreVariables(node->analysis.livevars)))));
	  statements = AppendItem(statements,
			       MakeCall(MakeId(AtThreadBoundarySlowName),
					NULL));
     }
     /* make block of stuff to replace spawn statement */
     result = MakeBlock(NULL, declarations, statements);
     SetCoords(result, coord, Subtree);
     return result;
}

PRIVATE inline Node *TransformSync(Node *node, syncNode *u)
{
     Node *result;
     List *statements;
     List *outerstatements;

     assert(in_procedure && !in_inlet);

     /* reserve a sync number */
     sync_count++;

     if (version == Fast) {
	  result = MakeCall(MakeId(AtSyncFastName), NULL);
	  SetCoords(result, node->coord, Subtree);
	  return result;
     }
     /* set up dummy link information */
     if (version == Slow) {
	  link_initializers =
	      AppendItem(link_initializers,
			 MakeInitializer(List5(MakeConstSint(0),
					       MakeConstSint(0),
					       MakeConstSint(0),
					       MakeConstSint(0),
					       MakeConstSint(0))));
     }
     /* 
      * generate sync code.  Sync code goes something like this:   
      * { 
      *     _BEFORE_SYNC_SLOW(); 
      *     _frame->_header.entry = 3;
      *     <save dirty variables> 
      *     if (_SYNC) { 
      *          _POP_FRAME(_frame); 
      *          return; 
      *    _sync3:; 
      *          <restore live variables> 
      *    } 
      *    _AFTER_SYNC_SLOW();
      * } 
      */

     outerstatements = List1(MakeCall(MakeId(BeforeSyncSlowName), NULL));

     /* set entry */
     outerstatements =
	  AppendItem(outerstatements,
		     MakeBinop('=',
			       MakeBinop('.',
					 MakeBinop(ARROW,
						   MakeId(FrameName),
						   MakeId(HeaderName)),
					 MakeId(EntryName)),
			       MakeConstSint(sync_count)));

     /* save live, dirty variables */
     outerstatements = JoinLists(outerstatements,
			  SaveVariables(node->analysis.dirtyvars, NULL));

     /* return */
     statements = List1(MakeReturn(NULL));

     /* sync label */
     statements = AppendItem(statements,
			     MakeLabel(MakeSyncName(sync_count),
				       NULL));

     /* restore variables */
     statements = JoinLists(statements,
			    RestoreVariables(node->analysis.livevars));

     outerstatements = AppendItem(outerstatements,
				  MakeIf(MakeId(SynchronizeName),
				     MakeBlock(NULL, NULL, statements)));

     outerstatements = AppendItem(outerstatements,
			      MakeCall(MakeId(AfterSyncSlowName), NULL));
     outerstatements = AppendItem(outerstatements,
			       MakeCall(MakeId(AtThreadBoundarySlowName),
					NULL));

     result = MakeBlock(NULL, NULL, outerstatements);

     SetCoords(result, node->coord, Subtree);
     return result;
}

PRIVATE Node *TransformInletCall(Node *node, inletcallNode *u)
{
     List *declarations = NULL;
     List *statements = NULL;
     Node *result;
     Node *spawnarg;
     List *otherargs;
     List *formals, *actuals;
     Node *temp_arg, *temp_decl;
     Node *index, *assignment;
     Coord coord = u->receiver ? u->receiver->coord : node->coord;
     Bool void_spawn;

     assert(in_procedure && !in_inlet);

     /* get spawn argument and any other arguments */
     assert(u->args != NULL);
     spawnarg = FirstItem(u->args);

     assert(spawnarg->typ == Spawn);
     void_spawn = IsVoidType(spawnarg);
     otherargs = Rest(u->args);

     u->receiver =
	  TransformReceiver(u->receiver, '=', NodeDataType(spawnarg), &index,
			    &assignment, &declarations, &statements);
     u->name = TransformNode(u->name);
     otherargs = TransformSpawnArgs(otherargs, &declarations, &statements);

     /* set up link information */
     if (version == Slow) {
	  Node *size, *inlet, *argindex, *argsize;
	  const char *scope;
	  const char *name;

	  /* calculate size entry - size of spawn's return value */
	  if (void_spawn)
	       size = MakeConstSint(0);
	  else
	       size = MakeUnary(SIZEOF, NodeDataType(spawnarg));

	  /* make inlet pointer */
	  name = MakeInletName(proc_name, u->name->u.id.decl->u.decl.name);
	  inlet = MakeId(name);

	  /* make argsize and argindex */
	  formals = GetSpawnFdcl(u->name)->u.fdcl.args;
	  if (formals != NULL && !IsVoidType(FirstItem(formals))) {
	       scope = ((Node *) FirstItem(formals))->u.decl.scope;
	       assert(scope != NULL);
	       argsize = MakeUnary(SIZEOF,
				   MakeInletArgsType(name));
	       argindex = MakeCall(MakeId(OffsetofName),
				   List2(MakeFrameType(proc_name),
					 MakeId(scope)));
	  } else {
	       argsize = MakeConstSint(0);
	       argindex = MakeConstSint(0);
	  }
	  link_initializers =
	      AppendItem(link_initializers,
			 MakeInitializer(List5(size,
					       index,
					       inlet,
					       argsize,
					       argindex)));
     }
     /* save other arguments besides spawn argument to frame */
     actuals = otherargs;
     if (void_spawn)
	  formals = GetSpawnFdcl(u->name)->u.fdcl.args;
     else
	  formals = Rest(GetSpawnFdcl(u->name)->u.fdcl.args);
     for (; actuals != NULL; actuals = Rest(actuals), formals = Rest(formals)) {
	  Node *actual = FirstItem(actuals);
	  Node *formal = FirstItem(formals);
	  formal = MakeBinop('.',
			     MakeBinop(ARROW,
				       MakeId(FrameName),
				       MakeId(formal->u.decl.scope)),
			     MakeId(formal->u.decl.name));
	  statements = AppendItem(statements,
				  MakeBinop('=',
					    formal,
					    actual));
     }

     /* transform spawn argument */
     if (!void_spawn) {
	  temp_decl = MakeTempDecl(NodeDataType(spawnarg));

	  /* treat this variable as a register variable, regardless of type */
	  temp_decl->u.decl.kind = 1;
	  declarations = AppendItem(declarations,
				    temp_decl);

	  temp_arg = MakeId(temp_decl->u.decl.name);
	  temp_arg->u.id.decl = temp_decl;

	  spawnarg->u.spawn.receiver = temp_arg;
	  otherargs = ConsItem(temp_arg, otherargs);
     } else
	  spawnarg->u.spawn.receiver = NULL;

     TransformSpawn2(spawnarg, &spawnarg->u.spawn, &declarations, &statements);

     /* do call of inlet */
     if (u->receiver)
	  statements = AppendItem(statements,
				  MakeBinop('=',
					    u->receiver,
					    MakeCall(u->name,
						     otherargs)));
     else
	  statements = AppendItem(statements,
				  MakeCall(u->name,
					   otherargs));

     /* save inlet result somewhere, if necessary */
     if (assignment != NULL)
	  statements = AppendItem(statements, assignment);

     /* Add entry recovery statements to slow version */
     if (version == Slow) {
	  statements = AppendItem(statements,
				  MakeIf(MakeConstSint(0),
					 MakeBlock(NULL, NULL,
			     ConsItem(MakeLabel(MakeSyncName(sync_count),
						NULL),
			   RestoreVariables(node->analysis.livevars)))));
	  statements = AppendItem(statements,
			       MakeCall(MakeId(AtThreadBoundarySlowName),
					NULL));
     }
     /* make block of stuff to replace inlet_call statement */
     result = MakeBlock(NULL, declarations, statements);
     SetCoords(result, coord, Subtree);
     return result;
}


PRIVATE inline Node *TransformAbort(Node *node, abortNode *u)
{
     Node *result;

     assert(in_procedure);

     switch (version) {
	 case Fast:
	      return MakeTextCoord(UniqueString("/* abort */;"), FALSE, node->coord);
	 case Slow:
	      result = MakeCall(MakeId(AbortSlowName), NULL);
	      SetCoords(result, node->coord, Subtree);
	      return result;
	 case InletProc:
	      result = MakeCall(MakeId(AbortStandaloneName), NULL);
	      SetCoords(result, node->coord, Subtree);
	      return result;
	 default:
	      FAIL("Unknown version");
     }

     return node;
}

PRIVATE inline Node *TransformSynched(Node *node, synchedNode *u)
{
     assert(in_procedure && !in_inlet);

     switch (version) {
	 case Fast:
	      return MakeConstSint(1);
	 case Slow:
	 case InletProc:
	      return MakeId(SynchedName);
	 default:
	      FAIL("Unknown version");
     }

     return node;
}

PRIVATE inline Node *TransformSpawnCounter(Node *node, spawncounterNode *u)
{
     assert(in_procedure && !in_inlet);
     assert(0);			/* removed spawn counter -KHR */

     switch (version) {
	 case Fast:
	      return MakeConstSint(0);
	 case Slow:
	 case InletProc:
	      return MakeId(SynchronizeName);
	 default:
	      FAIL("Unknown version");
     }

     return node;
}

/* if implicit sync is needed, calculating return expressions and
 * saving the result into temporary variables before implicit sync.
 * This is done first in TransformCilkProc so that the temporary
 * variables will be included in the procedure frames. -fengmd */
PRIVATE Node *AnalyzeReturnExpressions(Node *node)
{

#define CHILD(n) n = AnalyzeReturnExpressions(n)
     {
	  ASTWALK(node, CHILD)
     }
#undef CHILD

     if ((node->typ == Return) &&
	 (node->u.Return.needs_sync) &&
	 (node->u.Return.expr)) {
	  const char *tmp = MakeTempName();
	  Node *returntype = node->u.Return.proc->u.proc.decl->u.decl.type->u.fdcl.returns;
	  Node *decl = MakeDecl(tmp,
				T_BLOCK_DECL,
				NodeCopy(returntype, Subtree),
				node->u.Return.expr,
				NULL);
	  Node *tmp_id = MakeIdCoord(tmp, node->coord);
	  Node *return_node = MakeReturnCoord(tmp_id, node->coord);
	  List *decls = List1(decl);
	  Node *result;

	  tmp_id->u.id.decl = decl;
	  return_node->u.Return.needs_sync = 1;

	  /* set analysis results of return node for later passing to sync node */
	  return_node->u.Return.livevars = decls;
	  return_node->u.Return.dirtyvars = decls;

	  result = MakeBlockCoord(NULL, decls, List1(return_node),
				  node->coord, UnknownCoord);
	  SetCoords(result, node->coord, Subtree);
	  return result;
     } else
	  return node;
}


PRIVATE Node *MakeSlowDecl(const char *name)
{
     Node *arg = MakeDecl(FrameName,
			  T_FORMAL_DECL,
			  MakePtr(EMPTY_TQ, MakeFrameType(name)),
			  NULL,
			  NULL);

     Node *ws = MakeDecl(WorkerStateName,
			 T_FORMAL_DECL,
			 MakePtr(T_CONST, MakeWorkerStateType()),
			 NULL,
			 NULL);

     Node *n = 
	  MakeDecl(MakeSlowProcName(name),
		   T_STATIC | T_TOP_DECL,
		   MakeFdcl(EMPTY_TQ, List2(ws, arg), 
			    MakePrim(EMPTY_TQ, Void)),
		   NULL,
		   NULL);
     return n;
}

PRIVATE void SetDecl(Node *decl, List *register_vars, List *spawn_vars,
		     int scope_count)
{
     TypeQual sc = NodeStorageClass(decl);
     if (sc == T_EXTERN || sc == T_STATIC || sc == T_TYPEDEF ||
	 decl->u.decl.type->typ == Fdcl)
	  return;		/* leave as kind 0 */

     if (scope_count > 0 && DECL_LOCATION(decl->u.decl.tq) == T_FORMAL_DECL) {
	  /* inlet argument variable */
	  decl->u.decl.kind = 4;
     } else if (FindItem(register_vars, decl)) {
	  /* decl is for a register variable. */
	  if (FindItem(spawn_vars, decl))
	       decl->u.decl.kind = 3;
	  else
	       decl->u.decl.kind = 1;
     } else {
	  /* decl is for a frame-only variable. */
	  decl->u.decl.kind = 2;
     }

     decl->u.decl.scope = MakeScopeName(scope_count);
}

PRIVATE List *FindAllScopes(Node *node, List *declscopes)
{
     if (node->typ == Block && node->u.Block.decl) {
	  declscopes = 
	       AppendItem(declscopes, node->u.Block.decl);
     } else if (node->typ == Proc && IsInletProc(node)) {
	  /* add function arguments to scope list */
	  Node *type = node->u.proc.decl->u.decl.type;
	  assert(type->typ == Fdcl);

	  declscopes = AppendItem(declscopes, type->u.fdcl.args);
	  return declscopes;
	  /* return keeps FindAllScopes from looking inside local
	   * procedure */
     }
     /* transform children */
#define CHILD(n)  declscopes = FindAllScopes(n, declscopes)
     ASTWALK(node, CHILD)
#undef CHILD

     return declscopes;
}

PRIVATE List *FindSideEffectedVariables(Node *node, List *sofar)
{
     switch (node->typ) {
	 case Binop:
	      if (IsAssignmentOp(node->u.binop.op) &&
		  node->u.binop.left->typ == Id)
		   sofar = ConsItem(node->u.binop.left->u.id.decl, sofar);
	      break;
	 case Unary:
	      if (IsIncDecOp(node->u.unary.op) &&
		  node->u.unary.expr->typ == Id)
		   sofar = ConsItem(node->u.unary.expr->u.id.decl, sofar);
	      break;
	 default:
	      break;
     }

#define CODE(n)   sofar = FindSideEffectedVariables(n, sofar)
     ASTWALK(node, CODE)
#undef CODE

	 return sofar;
}

PRIVATE List *FindSpawnVariables(Node *node, List *sofar)
{
     if (node->typ == Spawn) {
	  if (node->u.spawn.receiver &&
	      node->u.spawn.receiver->typ == Id)
	       return ConsItem(node->u.spawn.receiver->u.id.decl, sofar);
     } else if (node->typ == InletCall) {
	  if (node->u.inletcall.receiver &&
	      node->u.inletcall.receiver->typ == Id)
	       return ConsItem(node->u.inletcall.receiver->u.id.decl, sofar);
     } else if (node->typ == Proc && IsInletProc(node)) {
	  /* need to find all variables modified by the inlet.  These
	   * variables need to be in the frame for the slow version. */
	  return JoinLists(sofar,
			   FindSideEffectedVariables(node, NULL));
     }
     /* transform children */
#define CHILD(n)  sofar = FindSpawnVariables(n, sofar)
     ASTWALK(node, CHILD)
#undef CHILD

	 return sofar;
}

PRIVATE void AnalyzeVariables(Node *node)
{
     ListMarker marker;
     Node *decl;
     int scope_count;

     /* find out which variables can be stored in registers */
     List *register_vars = RegisterVariables(node, NULL);
     List *spawn_vars = FindSpawnVariables(node, NULL);

     /* find all local scopes */
     List *scopes = FindAllScopes(node, NULL);
     scopes = ConsItem(node->u.proc.decl->u.decl.type->u.fdcl.args, scopes);

     /* iterate over all local variables in the procedure */
     scope_count = 0;
     while (scopes) {
	  IterateList(&marker, FirstItem(scopes));
	  while (NextOnList(&marker, (GenericREF) & decl)) {
	       if (decl->typ == Decl)
		    SetDecl(decl, register_vars, spawn_vars, scope_count);
	  }
	  scopes = Rest(scopes);
	  scope_count++;
     }
}

PRIVATE Node *ExpandTypedefs(Node *node)
{
     Node *junk;
     Node *new = NodeCopy(node, NodeOnlyExact);

#define CHILD(n) n = ExpandTypedefs(n)
     {
	  ASTWALK(new, CHILD)
     }
#undef CHILD

     /* expand locally declared typedefs */
     if (new->typ == Tdef) {
	  if (LookupSymbol(Identifiers, new->u.tdef.name,
			   (GenericREF) & junk)) {
	       if (NodeStorageClass(junk) == T_TYPEDEF &&
		   junk->u.decl.type == new->u.tdef.type)
		    return new;	/* external declaration is OK */
	  }
	  /*
	   * external declaration is not the same as the local one - must
	   * expand the typedef. 
	   */
	  /* 
	   * tdef.type is not visited when doing an ASTWALK! could we
	   * get into circularity problems?  I don't think so.  -KHR 
	   */
	  new->u.tdef.type = ExpandTypedefs(new->u.tdef.type);

	  /*
	   * is this the right thing to do with tdef's type qualifiers? -
	   *  KHR 
	   */
	  NodeAddTq(new->u.tdef.type, new->u.tdef.tq);
	  return new->u.tdef.type;
     } else
	  return new;
}

PRIVATE Node *MakeFrame(procNode *u)
{
     List *scopes, *scope;
     int scope_count;
     List *fields;
     ListMarker marker1, marker2;
     Node *result;
     Node *type;

     /* 
      * make frame:
      *
      * typedef struct foo_frame {
      *   StackFrameT _header;
      *   { ... args ... } _scope0;
      *   { ... scope 1 vars ...} _scope1;
      *   ...
      *   { ... scope <n> vars ...} _scope<n>;
      * }
      *
      * for each different type of inlet, we also have a scope that
      * holds its argument list.
      * { ... args of inlet 1 ... } _scope<x>;
      * { ... args of inlet <n> ... } _scope<y>;
      *
      * The inlet scopes are mixed in with the regular scopes in no
      * particular order.  We could put all of the inlet argument
      * scopes in a oneof because only one is being used at a time.
      */
     fields = List1(MakeDecl(HeaderName,
			     EMPTY_TQ | T_SU_DECL,
			     MakeGenericFrameType(),
			     NULL,
			     NULL));

     /* if proc return type is not an arithmetic type, put it in frame */
     type = u->decl->u.decl.type->u.fdcl.returns;
     if (!IsVoidType(type) && !IsScalarType(type)) {
	  fields = AppendItem(fields,
			      MakeDecl(DummyReturnName,
				       EMPTY_TQ | T_SU_DECL,
				       NodeCopy(type, Subtree),
				       NULL,
				       NULL));
     }
     /* add scopes to frame */
     scopes = FindAllScopes(u->body, NULL);
     scopes = ConsItem(u->decl->u.decl.type->u.fdcl.args, scopes);

     scope_count = 0;
     IterateList(&marker1, scopes);
     while (NextOnList(&marker1, (GenericREF) & scope)) {
	  List *scopefields = NULL;
	  Node *decl;

	  IterateList(&marker2, scope);
	  while (NextOnList(&marker2, (GenericREF) & decl)) {
	       if (decl->typ != Decl)
		    continue;
	       if (decl->u.decl.kind == 0)
		    continue;
	       SetCurrentOnList(&marker2, NodeCopy(decl, Subtree));
	       NodeRemoveTq(decl, T_REGISTER);
	       NodeRemoveTq(decl->u.decl.type, T_CONST);
	       decl->u.decl.type = ExpandTypedefs(decl->u.decl.type);
	       decl->u.decl.init = NULL;
	       decl = TransformNode(NodeCopy(decl, Subtree));
	       scopefields = AppendItem(scopefields, decl);
	  }

	  if (scopefields)
	       fields = AppendItem(fields,
				   MakeDecl(MakeScopeName(scope_count),
					    EMPTY_TQ | T_SU_DECL,
				    MakeSdcl(EMPTY_TQ | T_SUE_ELABORATED,
				      make_SUE(Sdcl, NULL, scopefields)),
					    NULL,
					    NULL));
	  scope_count++;
     }

     /* finally, make struct declaration for frame */
     result = MakeDecl(NULL,
		       EMPTY_TQ | T_TOP_DECL,
		       MakeSdcl(EMPTY_TQ | T_SUE_ELABORATED,
	make_SUE(Sdcl, MakeFrameTypeName(u->decl->u.decl.name), fields)),
		       NULL,
		       NULL);
     SetCoords(result, UnknownCoord, Subtree);
     return result;
}

PRIVATE inline Node *MakeLinkage(Node *node)
{
     Node *linkage;

     /* make CilkProcInfo declaration */
     linkage = MakeDecl(MakeLinkName(proc_name),
			T_STATIC,
			MakeAdcl(EMPTY_TQ,
				 MakeTdef(EMPTY_TQ,
					  UniqueString("CilkProcInfo")),
				 NULL),
			MakeInitializer(link_initializers),
			NULL);
     SetCoords(linkage, UnknownCoord, Subtree);
     return (linkage);
}

/*
 * Add code to allocate frame to beginning of fast body.  For example:
 *    _foo_frame *_frame;
 *    _INIT_FRAME(_frame, sizeof(_foo_frame), _foo_sig);
 *    _START_THREAD_FAST();
 *    (_INIT_FRAME does, among other things,
 *           _frame->_header.size = sizeof(_foo_frame);
 *           _frame->_header.sig = _foo_sig; )
 *
 *    ... transfer non-register arguments to frame ...
 */
PRIVATE Node *AddFastStuff(Node *body, List *formals)
{
     List *stmts;
     ListMarker marker;
     Node *decl;

     stmts = List1(MakeCall(MakeId(InitFrameName),
			    List3(MakeId(FrameName),
				  MakeUnary(SIZEOF,
					    MakeFrameType(proc_name)),
				  MakeId(MakeLinkName(proc_name)))));

     stmts = AppendItem(stmts,
			MakeCall(MakeId(StartThreadFastName), NULL));

     /* move any frame-only formals to frame */
     IterateList(&marker, formals);
     while (NextOnList(&marker, (GenericREF) & decl)) {
	  if (decl->u.decl.kind == 2) {
	       const char *name = decl->u.decl.name;
	       const char *scope = decl->u.decl.scope;

	       /* 
	        * we need to do something different for array arguments
	        * -KHR 
	        */
	       stmts = AppendItem(stmts,
				  MakeBinop('=',
					    MakeBinop('.',
						      MakeBinop(ARROW,
						       MakeId(FrameName),
							  MakeId(scope)),
						      MakeId(name)),
					    MakeId(name)));
	  }
     }

     /* append code of procedure here */
     stmts = AppendItem(stmts, body);

     return MakeBlock(NULL,
		      List1(MakeDecl(FrameName,
				     EMPTY_TQ | T_BLOCK_DECL,
				     MakePtr(EMPTY_TQ,
					     MakeFrameType(proc_name)),
				     NULL,
				     NULL)),
		      stmts);
}

/* 
 * Add case statement and argument unpacking to beginning of slow body.
 * For example:
 *    int n;       < declaration for each register formal >
 *    _START_THREAD_SLOW();
 *    switch (_frame->_header.entry) {
 *      case 1: goto _sync1;
 *      case 2: goto _sync2;
 *      case 3: goto _sync3;
 *    }
 *    n = _frame->n;   < initialization for each register formal >
 */
Node *AddSlowStuff(Node *body, List *formals)
{
     List *cases = NULL;
     List *regdecls = NULL;
     ListMarker marker;
     Node *decl;

     /* make cases for case statement */
     for (; sync_count; sync_count--) {
	  cases = ConsItem(MakeCase(MakeConstSint(sync_count),
			     MakeGoto(MakeLabel(MakeSyncName(sync_count),
						NULL)),
				    NULL),
			   cases);
     }

     /* get declarations for register formals */
     IterateList(&marker, formals);
     while (NextOnList(&marker, (GenericREF) & decl)) {
	  if (decl->u.decl.kind == 1)
	       regdecls = AppendItem(regdecls, decl);
     }

     return MakeBlock(NULL,
		      regdecls,
		      ConsItem(
			     MakeCall(MakeId(StartThreadSlowName), NULL),
				   ConsItem(
						MakeSwitch(MakeBinop('.',
							 MakeBinop(ARROW,
						       MakeId(FrameName),
						     MakeId(HeaderName)),
						      MakeId(EntryName)),
					    MakeBlock(NULL, NULL, cases),
							   cases),
				   AppendItem(RestoreVariables(regdecls),
					      body))));
}

Node *GlobalizeSUE(SUEtype *type)
{
     Node *junk;
     assert(type->complete);

     /* choose a globally unique name for this SUE */
     if (LookupSymbol(Tags, type->name, (GenericREF) & junk)) {
	  char buf[40];
	  sprintf(buf, "%s%.16s", CilkPrefix, type->name);
	  type->name = InsertUniqueSymbol(Tags, type, buf);
     }
     switch (type->typ) {
	 case Sdcl:
	      return MakeSdcl(T_SUE_ELABORATED, type);
	 case Udcl:
	      return MakeUdcl(T_SUE_ELABORATED, type);
	 case Edcl:
	      return MakeEdcl(T_SUE_ELABORATED, type);
	 default:
	      FAIL("bad type in GlobalizeSUE");
	      return NULL;
     }
}

List *MoveSUEToTop(Node *node, List *sofar)
{
     Node *new = NULL;

#define CODE(n)   sofar = MoveSUEToTop(n, sofar)
     ASTWALK(node, CODE)
#undef CODE

	 switch (node->typ) {
	 case Sdcl:
	      if (SUE_ELABORATED(node->u.sdcl.tq)) {
		   new = GlobalizeSUE(node->u.sdcl.type);
		   NodeRemoveTq(node, T_SUE_ELABORATED);
	      }
	      break;
	 case Udcl:
	      if (SUE_ELABORATED(node->u.udcl.tq)) {
		   new = GlobalizeSUE(node->u.udcl.type);
		   NodeRemoveTq(node, T_SUE_ELABORATED);
	      }
	      break;
	 case Edcl:
	      if (SUE_ELABORATED(node->u.edcl.tq)) {
		   new = GlobalizeSUE(node->u.edcl.type);
		   NodeRemoveTq(node, T_SUE_ELABORATED);
	      }
	      break;
	 case Decl:
	      if (NodeStorageClass(node) == T_STATIC &&
		  NodeDeclLocation(node) == T_BLOCK_DECL) {
		   Node *junk;
		   /* choose a globally unique name for this static *
		    * variable */
		   if (LookupSymbol(Identifiers, node->u.decl.name, (GenericREF) & junk)) {
			node->u.decl.name = InsertUniqueSymbol(Identifiers, node, node->u.decl.name);
		   } else {
			InsertSymbol(Identifiers, node->u.decl.name, node, NULL);
		   }
		   new = node;
		   NodeSetDeclLocation(new, T_TOP_DECL);
		   new = TransformDecl(new, &new->u.decl);

		   /* 
		    * Note: the static variables will be renamed (if necessary)
		    * by TransformId, and the local static declaration will be
		    * removed by TransformDecl.    -KHR
		    */
	      }
	      break;
	 default:
	      break;
     }

     if (new) {
	  /*
	   * make sure we expand any typedefs of declarations moved to 
	   * the top level.   -KHR 
	   */
	  new = ExpandTypedefs(new);
	  SetCoords(new, UnknownCoord, Subtree);
	  sofar = AppendItem(sofar, new);
     }
     return sofar;
}

PRIVATE inline List *TransformCilkProc(Node *node, procNode *u)
{
     Node *fastnode, *slownode;
     Node *decl;
     List *result;

     /* if implicit sync is needed, calculating return expressions and
      * saving the result into tempary variables before implicit sync.
      * -fengmd */
     node = AnalyzeReturnExpressions(node);

     /* analyze procedure variables */
     AnalyzeVariables(node);

     /* start with empty text node - just enforces coord start */
     result = List1(MakeTextCoord("", FALSE, node->coord));

     /* move structures/unions/enums, static variables to outer scope */
     result = JoinLists(result, MoveSUEToTop(node, NULL));

     /* make frame declaration for procedure */
     result = AppendItem(result, MakeFrame(u));

     /* write down some global variables for other routines to read */
     proc_name = u->decl->u.decl.name;
     curr_proc = node;

     fastnode = NodeCopy(node, Subtree);
     slownode = NodeCopy(node, Subtree);

     /* add implicit sync and return to slow and fast bodies, if needed */
     if (node->u.proc.needs_sync) {
	  slownode->u.proc.body->u.Block.stmts =
	      AppendItem(slownode->u.proc.body->u.Block.stmts,
	      MakeSyncCoord(slownode->u.proc.body->u.Block.right_coord));
	  fastnode->u.proc.body->u.Block.stmts =
	      AppendItem(fastnode->u.proc.body->u.Block.stmts,
	      MakeSyncCoord(fastnode->u.proc.body->u.Block.right_coord));
     }
     if (node->u.proc.needs_return) {
	  slownode->u.proc.body->u.Block.stmts =
	      AppendItem(slownode->u.proc.body->u.Block.stmts,
			 MakeReturnCoord(NULL,
			    slownode->u.proc.body->u.Block.right_coord));
	  fastnode->u.proc.body->u.Block.stmts =
	      AppendItem(fastnode->u.proc.body->u.Block.stmts,
			 MakeReturnCoord(NULL,
			    fastnode->u.proc.body->u.Block.right_coord));
     }
     /* transform fast body */
     sync_count = 0;
     version = Fast;
     fastnode->u.proc.decl = TransformNode(fastnode->u.proc.decl);
     in_procedure = 1;
     fastnode->u.proc.body = TransformNode(fastnode->u.proc.body);
     in_procedure = 0;

     /* transform slow body */
     sync_count = 0;
     version = Slow;
     /* add info for entry 0 */
     link_initializers =
	  List1(MakeInitializer(
	       List5(IsVoidType(u->decl->u.decl.type->u.fdcl.returns) ?
		     MakeConstSint(0) : 
		     MakeUnary(SIZEOF, NodeCopy(u->decl->u.decl.type->u.fdcl.returns, Subtree)),
		     MakeUnary(SIZEOF, MakeFrameType(proc_name)),
		     MakeId(MakeSlowProcName(proc_name)),
		     MakeConstSint(0),
		     MakeConstSint(0)
		    )
	       ));
     standalone_inlets = NULL;
     slownode->u.proc.decl = TransformNode(slownode->u.proc.decl);
     in_procedure = 1;
     slownode->u.proc.body = TransformNode(slownode->u.proc.body);
     in_procedure = 0;

     /* add stuff to beginning of procedure */
     fastnode->u.proc.body = AddFastStuff(fastnode->u.proc.body,
			fastnode->u.proc.decl->u.decl.type->u.fdcl.args);
     slownode->u.proc.body = AddSlowStuff(slownode->u.proc.body,
			slownode->u.proc.decl->u.decl.type->u.fdcl.args);

     /* make slow decl */
     /* 
      * cilk int foo(int x, int y)  ==>  void _foo_slow(_foo_frame *_frame)
      */
     decl = MakeSlowDecl(proc_name);
     SetCoords(decl, slownode->coord, Subtree);
     slownode->u.proc.decl = decl;

     /* make slow procedure declaration */
     result = AppendItem(result,
			 MakeSlowDecl(u->decl->u.decl.name));

     /* add in standalone inlets */
     result = JoinLists(result, standalone_inlets);

     /* add linkage stuff */
     result = AppendItem(result, MakeLinkage(node));

     result = AppendItem(result, fastnode);
     result = AppendItem(result, slownode);

     return result;
}

PRIVATE Node *TransformNode(Node *node)
{
     if (node == NULL)
	  return NULL;
#define CODE(name, node, union) return Transform##name(node, union)
     ASTSWITCH(node, CODE)
#undef CODE
	 UNREACHABLE;
     return NULL;
}

GLOBAL List *TransformTopDecl(Node *node)
{
     List *result;

     if (!node)
	  return NULL;

     if (node->typ == Proc &&
	 node->u.proc.decl->u.decl.type->typ == Fdcl &&
	 node->u.proc.decl->u.decl.type->u.fdcl.tq & T_PROCEDURE) {
	  /* cilk procedure */
	  result = TransformCilkProc(node, &node->u.proc);
     } else
	  result = List1(TransformNode(node));

     return (result);
}

GLOBAL List *TransformPass1(List *program)
{
     ListMarker marker;
     Node *node;
     List *newprogram = NULL;

     IterateList(&marker, program);
     while (NextOnList(&marker, (GenericREF) & node)) {
	  newprogram = JoinLists(newprogram, TransformTopDecl(node));
     }

     return newprogram;
}

/*
 * TransformProgram should convert a type-checked source language tree
 * into a standard C tree.  Type information does not need to be preserved,
 * but all new Decl nodes must be properly annotated with their DECL_LOCATION
 * (e.g. top-level, block, structure field, etc.).
 */

GLOBAL List *TransformProgram(List *program)
{
     InitCilkTransform();
     program = TransformPass1(program);
     program = ConsItem(MakeText("#include <cilk-cilk2c.h>\n", TRUE),
			program);
     return program;
}
